package dev.fitko.fitconnect.client.sender.strategies;

import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;

import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSentSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmissionToAnnounce;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmitSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.toAttachmentPayloads;

public class SendEncryptedSubmissionStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendEncryptedSubmissionStrategy.class);

    private final Sender sender;

    public SendEncryptedSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    public SentSubmission send(final EncryptedSubmissionPayload submissionPayload) {

        try {
            final UUID submissionId = announceNewSubmission(submissionPayload);
            final SubmitSubmission submitSubmission = buildSubmitSubmission(submissionId, submissionPayload);
            final List<AttachmentPayload> attachmentPayloads = toAttachmentPayloads(submissionPayload.getEncryptedAttachments());

            final var startTimeAttachmentUpload = StopWatch.start();
            uploadAttachments(attachmentPayloads, submissionId);
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));

            final var startTimeSubmissionUpload = StopWatch.start();
            final Submission submission = sender.sendSubmission(submitSubmission);
            LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            return buildSentSubmission(submissionPayload, attachmentPayloads, submission);

        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        }

        return null;
    }

    private UUID announceNewSubmission(final EncryptedSubmissionPayload submissionPayload) {
        final CreateSubmission submissionToAnnounce = buildSubmissionToAnnounce(submissionPayload);
        return sender.createSubmission(submissionToAnnounce).getSubmissionId();
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
            attachmentPayloads.forEach(a -> sender.uploadAttachment(submissionId, a.getAttachmentId(), a.getEncryptedData()));
        }
    }
}
