package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.*;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.factory.ClientFactory;
import dev.fitko.fitconnect.client.sender.EncryptedSubmissionBuilder;
import dev.fitko.fitconnect.client.sender.SubmissionBuilder;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.client.util.SubmissionUtil;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import org.apache.tika.mime.MimeTypes;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.ParseException;
import java.time.Duration;
import java.util.*;
import java.util.stream.Collectors;

import static org.awaitility.Awaitility.await;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.hasItems;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;

/**
 * This test uses real credentials and endpoints stored in gitlab ci variables to test the sdk against a real system
 */
class ClientIntegrationTest {

    private static final String authBaseUrl = "https://auth-testing.fit-connect.fitko.dev";
    private static final String submissionBaseUrl = "https://submission-api-testing.fit-connect.fitko.dev";

    @Nested
    class SendSubmissionTests {

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testSendAndConfirmCycle() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final ReceivedSubmission receivedSubmission =
                    ClientFactory.subscriberClient(config)
                            .requestSubmission(sentSubmission.getSubmissionId());

            // Then
            assertNotNull(receivedSubmission);
            assertThat(receivedSubmission.getData(), is("{ \"data\": \"Beispiel Fachdaten\" }"));
            assertThat(receivedSubmission.getDataSchemaUri(), is(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")));
            assertThat(receivedSubmission.getDataMimeType().toString(), is("application/json"));
            assertThat(new String(receivedSubmission.getAttachments().get(0).getData()), is("Test attachment"));
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testSendAndConfirmCycleWithEncryptedData() throws ParseException, IOException {

            // Given
            final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);
            final CryptoService cryptoService = new JWECryptoService(new HashService());
            final SenderClient client = ClientFactory.senderClient(config);

            final Optional<String> publicKey = client.getPublicKey(destinationId);
            assertTrue(publicKey.isPresent());

            final RSAKey encryptionKey = RSAKey.parse(publicKey.get());

            final String jsonData = "{ \"data\": \"Beispiel Fachdaten\" }";
            final File attachmentFile = new File("src/test/resources/attachment.txt");
            final String attachmentData = Files.readString(attachmentFile.toPath());

            final String encryptedData = cryptoService.encryptString(encryptionKey, jsonData);
            final String encryptedAttachment = cryptoService.encryptString(encryptionKey, attachmentData);

            final SubmissionSchema submissionSchema = new SubmissionSchema();
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

            final Hash hash = new Hash();
            hash.setSignatureType(SignatureType.SHA_512);
            hash.setContent(cryptoService.hashBytes(jsonData.getBytes(StandardCharsets.UTF_8)));

            final Data data = new Data();
            data.setHash(hash);
            data.setSubmissionSchema(submissionSchema);

            final AttachmentPayload attachmentPayload = AttachmentPayload.builder()
                    .encryptedData(encryptedAttachment)
                    .file(attachmentFile)
                    .mimeType(MimeTypes.PLAIN_TEXT)
                    .attachmentId(UUID.randomUUID())
                    .hashedData(cryptoService.hashBytes(attachmentData.getBytes(StandardCharsets.UTF_8)))
                    .build();

            final ServiceType serviceType = ServiceType.builder()
                    .name("Test Service")
                    .identifier("urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final PublicServiceType publicServiceType = SubmissionUtil.buildPublicServiceType(serviceType);
            final List<Attachment> attachmentMetadata = SubmissionUtil.toAttachmentMetadata(List.of(attachmentPayload));
            final Metadata metadata = SubmissionUtil.buildMetadata(attachmentMetadata, data, publicServiceType);

            final String encryptedMetadata = cryptoService.encryptBytes(encryptionKey, new ObjectMapper().writeValueAsBytes(metadata));

            // When
            final var submission = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedAttachment(attachmentPayload.getAttachmentId(), attachmentPayload.getEncryptedData())
                    .withEncryptedData(encryptedData)
                    .withEncryptedMetadata(encryptedMetadata)
                    .withDestination(destinationId)
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            final ReceivedSubmission receivedSubmission = ClientFactory.subscriberClient(config).requestSubmission(sentSubmission.getSubmissionId());

            // Then
            assertNotNull(receivedSubmission);
            assertThat(receivedSubmission.getData(), is(jsonData));
            assertThat(receivedSubmission.getDataMimeType().toString(), is("application/json"));
            assertThat(new String(receivedSubmission.getAttachments().get(0).getData()), is("Test attachment"));
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testAbortedSendSubmissionWithKeyValidationNotSilent() {

            // Given
            final var prodEnv = new Environment(authBaseUrl, submissionBaseUrl, false);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("PROD", prodEnv);

            // When
            final var submission = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            // Then
            assertNull(sentSubmission);
        }
    }

    @Nested
    class ReceiveSubmissionTests {

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testListSubmissions() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var senderClient = ClientFactory.senderClient(config);
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
            final String leikaKey = "urn:de:fim:leika:leistung:99400048079000";
            final String serviceName = "Test Service";

            final var submissionOne = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten 1\" }")
                    .withDestination(destinationId)
                    .withServiceType(serviceName, leikaKey)
                    .build();

            final var submissionTwo = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten 2\" }")
                    .withDestination(destinationId)
                    .withServiceType(serviceName, leikaKey)
                    .build();

            final var sentSubmissionOne = senderClient.submit(submissionOne);
            final var sentSubmissionTwo = senderClient.submit(submissionTwo);

            assertNotNull(sentSubmissionOne);
            assertNotNull(sentSubmissionTwo);

            // When
            final Set<SubmissionForPickup> submissions = subscriberClient.getAvailableSubmissions(destinationId);

            // Then
            assertFalse(submissions.isEmpty());

            final List<UUID> submissionIds = submissions.stream().map(SubmissionForPickup::getSubmissionId).collect(Collectors.toList());

            assertThat(submissionIds, hasItems(sentSubmissionOne.getSubmissionId(), sentSubmissionTwo.getSubmissionId()));

            // remove by confirming
            subscriberClient.requestSubmission(sentSubmissionOne.getSubmissionId()).acceptSubmission();
            subscriberClient.requestSubmission(sentSubmissionTwo.getSubmissionId()).acceptSubmission();
        }
    }

    @Nested
    class EventTests {

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testRejectEvent() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final var sentSubmissionId = sentSubmission.getSubmissionId();

            // reject and remove
            subscriberClient.requestSubmission(sentSubmissionId).rejectSubmission(List.of(new InvalidEventLog()));

            // second attempt to receive and reject the submission should return an empty result
            // since the submission is gone after being rejected
            assertNull(subscriberClient.requestSubmission(sentSubmissionId));
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testAcceptEvent() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final var sentSubmissionId = sentSubmission.getSubmissionId();

            // accept and remove
            subscriberClient.requestSubmission(sentSubmissionId).acceptSubmission();

            // second attempt to receive the submission should return an empty result
            // since the submission is gone after being accepted
            assertNull(subscriberClient.requestSubmission(sentSubmissionId));
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testReadEventLogFromSender() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var submission = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                final UUID destinationId = sentSubmission.getDestinationId();
                final UUID caseId = sentSubmission.getCaseId();

                final List<EventLogEntry> senderEventLog = ClientFactory.senderClient(config).getEventLog(caseId, destinationId);
                final List<Event> senderEvents = senderEventLog.stream().map(EventLogEntry::getEvent).collect(Collectors.toList());

                // Then
                return senderEvents.equals(List.of(Event.CREATE, Event.SUBMIT, Event.NOTIFY));
            });
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testReadEventLogFromSubscriber() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            final var receivedSubmission = ClientFactory.subscriberClient(config)
                    .requestSubmission(sentSubmission.getSubmissionId());

            assertNotNull(receivedSubmission);

            // When
            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                final UUID destinationId = receivedSubmission.getDestinationId();
                final UUID caseId = receivedSubmission.getCaseId();

                final List<EventLogEntry> subscriberEventLog = ClientFactory.subscriberClient(config).getEventLog(caseId, destinationId);
                final List<Event> subscriberEvents = subscriberEventLog.stream().map(EventLogEntry::getEvent).collect(Collectors.toList());

                // Then
                return subscriberEvents.equals(List.of(Event.CREATE, Event.SUBMIT, Event.NOTIFY));
            });
        }

        @Test
        @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
        void testReadSubmissionStatusWithAuthTagEventValidationFromSender() {

            // Given
            final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

            final SenderClient senderClient = ClientFactory.senderClient(config);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                // When
                final EventStatus statusForSubmission = senderClient.getStatusForSubmission(sentSubmission);

                // Then
                return statusForSubmission.getStatus().equals(Event.SUBMIT.getState());
            });
        }
    }


    private ApplicationConfig getConfigWithCredentialsFromGitlab(final Environment env) {
        return getConfigWithCredentialsFromGitlab("TESTING", env);
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab(final String environmentName, final Environment env) {

        final var sender = new SenderConfig(System.getenv("SENDER_CLIENT_ID"), System.getenv("SENDER_CLIENT_SECRET"));

        final var subscriber = SubscriberConfig.builder()
                .clientId(System.getenv("SUBSCRIBER_CLIENT_ID"))
                .clientSecret(System.getenv("SUBSCRIBER_CLIENT_SECRET"))
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var envName = new EnvironmentName(environmentName);

        return ApplicationConfig.builder()
                .senderConfig(sender)
                .subscriberConfig(subscriber)
                .environments(Map.of(envName, env))
                .activeEnvironment(envName)
                .build();
    }
}
